/*=================================================================
 * The main routine analyzes all incoming (right-hand side) arguments 
 *
 * Copyright 1984-2006 The MathWorks, Inc.
 * $Revision: 1.14.4.5 $ 
 *	
 *=================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mex.h"

#include <vector>

#include <gsl/gsl_rng.h>
#include <math.h>

#include "clock.h"

using namespace std;

void        display_subscript(const mxArray *array_ptr, mwSize index);
void        get_characteristics(const mxArray  *array_ptr);
mxClassID   analyze_class(const mxArray *array_ptr);

typedef unsigned int uint;
typedef double TReward;

class Params {
public:
    TReward rMax;
    double sampNum, gamma, tryNum, tol, dt;

    Params() {}

    Params( TReward _rMax, double _sampNum, double _gamma, double _tryNum, double _tol, double _dt )
    : rMax(_rMax), sampNum(_sampNum), gamma(_gamma), tryNum(_tryNum), tol(_tol), dt(_dt) {};

    void printf() const {
        ::mexPrintf("\trMax    = %.3f\n", rMax);
        ::mexPrintf("\tsampNum = %.3f\n", sampNum);
        ::mexPrintf("\tgamma   = %.3f\n", gamma);
        ::mexPrintf("\ttryNum  = %.3f\n", tryNum);
        ::mexPrintf("\ttol     = %.3e\n", tol);
        ::mexPrintf("\tdt      = %.3f\n", dt);
    }
};

class tA {
public:
    TReward r;
    vector<double> P;
    tA(const TReward &_r, const vector<double> &_P) : r(_r), P(_P) {};
};

class tS {
public:
    uint ASz;
    vector<tA> A;
    tS() {ASz=0;}
    tS(const uint &_ASz, const vector<tA> &_A) : ASz(_ASz), A(_A) {};
};

class MDP {
public:
    uint SSz;
    vector<tS> S;

    void printf() const {
        for (uint s=0; s<SSz; s++) {
            for (uint a=0; a<S[s].ASz; a++) {
                ::mexPrintf("%d: (r=%f)\t", a, S[s].A[a].r);
            }
            ::mexPrintf("\n");
        }
    }    
};

typedef vector<vector<double> > tQ;
typedef vector<vector<vector<double> > > tMdl;
typedef vector<vector<tQ> > tQFinal;
typedef vector<vector<vector<short int> > > sampIndex;

void initQ(const Params &p, const MDP &mdp, gsl_rng *r, tQ &Q) {
    for (uint s=0; s<mdp.SSz; s++) 
        for (uint a=0; a<mdp.S[s].ASz; a++) 
           Q[s][a] = (double)p.rMax*(2.0*gsl_rng_uniform(r)-1 )/(1.-p.gamma);
            
}

//void initQ(const Params &p, const MDP &mdp,  tQ &Q) {
//    for (uint s=0; s<mdp.SSz; s++) 
//        for (uint a=0; a<mdp.S[s].ASz; a++) 
//            Q[s][a] = (double)p.rMax*(2.0*rand()/(double)RAND_MAX-1 )/(1.-p.gamma);
//}

void printQ(const tQ &Q) {
    for (uint s=0; s<Q.size(); s++) {
        for (uint a=0; a<Q[s].size(); a++) {
            mexPrintf("%.5f ",Q[s][a]);
        }
        mexPrintf("\n");
    }
}

//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////Main Function////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////


void doIt(  const Params &p,
            const MDP &mdp,
            tQFinal &Qf,
            tQ &Qo,
            double *vit,
            double *vcpu)
{
    p.printf();
    // init random number generation
    gsl_rng * r;
    const gsl_rng_type * T;
    gsl_rng_env_setup();
    T = gsl_rng_default;
    r = gsl_rng_alloc (T);
    mexPrintf ("seed = %lu\n", gsl_rng_default_seed);
    mexEvalString("drawnow;");

    //srand ( time(NULL) );
    
    // initialize sizes of vectors
    tQ Q, QN, QP;
    tMdl Cdfs;
    tMdl Mdl;

    Q.resize(mdp.SSz); // is S x A
    QP.resize(mdp.SSz); 
    Mdl.resize(mdp.SSz);                // is S x A x S
    Cdfs.resize(mdp.SSz);               // is S x A x S
    QN.resize(mdp.SSz);                 // is S x A
    Qf.resize((uint)(p.tryNum));                // is tryNum x dt x S x A
    for (uint s=0; s<mdp.SSz; s++) {
        Q[s].resize(mdp.S[s].ASz);
        QP[s].resize(mdp.S[s].ASz);
        Mdl[s].resize(mdp.S[s].ASz);
        Cdfs[s].resize(mdp.S[s].ASz);
        QN[s].resize(mdp.S[s].ASz);
        for (uint a=0; a<QN[s].size(); a++) {
            Mdl[s][a].resize(mdp.SSz);
            Cdfs[s][a].resize(mdp.SSz);
        }
    }
    
    // calculate optimum
    // Init Q
    Clock clk0;
    initQ(p, mdp,  r, Q);  
    initQ(p, mdp,  r, QN);  
    initQ(p, mdp,  r, QP);
    //initQ(p, mdp,   Q);  
    //initQ(p, mdp,   QN); 
    double dq = 1;
    uint ito=0;
    while (dq > p.tol) {
        dq = 1e-14;
        for (uint s=0; s<mdp.SSz; s++) {
            for (uint a=0; a<mdp.S[s].ASz; a++) {
                double daux=QN[s][a];
                QN[s][a]=mdp.S[s].A[a].r;
                for (uint sp=0; sp<mdp.SSz; sp++) {
                    double maxQ = -p.rMax/(1.-p.gamma);
                    for (uint ap=0; ap<mdp.S[sp].ASz; ap++)  
                        if (Q[sp][ap] > maxQ) maxQ = Q[sp][ap];
                    QN[s][a] += p.gamma*mdp.S[s].A[a].P[sp]*maxQ;
                }
                double dif = fabs(QN[s][a]-daux);
                if (dif > dq) dq = dif;
            }
        }
        Q = QN;
        ito++;
    }

    mexPrintf("Opt in %.3f scnds\n", clk0.toc());
    mexEvalString("drawnow;");

    // make cdfs for samples
    clk0.tic();
    for (uint s=0; s<mdp.SSz; s++) {
        for (uint a=0; a<mdp.S[s].ASz; a++) {
            Cdfs[s][a][0] = mdp.S[s].A[a].P[0];
            for (uint v=0; v<(mdp.SSz-1); v++)
                Cdfs[s][a][v+1] = Cdfs[s][a][v]+mdp.S[s].A[a].P[v+1];
        }
    }
    mexPrintf("Cdfs computed in %.3f scnds\n", clk0.toc());
    mexEvalString("drawnow;");
   
   sampIndex indN;
   double rn;
   
  
    
    
    
    Qo=Q;
    
    uint ddt = (uint)floor(p.sampNum/p.dt);
    for (uint i=0; i<(uint)(p.tryNum); i++) {
       
    	// Init Q
        initQ(p, mdp, r, Q); 
        
        //initQ(p, mdp,  Q);  
        Qf[i].resize((uint)(p.dt)+2);
        uint its = 0;
        QN=Q;
        QP=Q;
        
        
        indN.resize(mdp.SSz);
        for (uint s=0; s<mdp.SSz; s++) {
         indN[s].resize(mdp.S[s].ASz);
         for (uint a=0; a<mdp.S[s].ASz; a++) {
             indN[s][a].resize((uint)p.sampNum+1);
             for (uint ns=0; ns<=(uint)p.sampNum; ns++) { 
                    
                    // generate one sample
                    rn=gsl_rng_uniform(r);
                    //rn=(double)rand()/(double)RAND_MAX;
                    
                    uint lowerbound=0,upperbound=mdp.SSz-1;
                    
                    //bool found = false;
                    
                    uint tempVal = upperbound/2;
                    
                
                    if (Cdfs[s][a][0]<=rn)
                    {
                    while ((upperbound-lowerbound)>1 ) {
                        if (Cdfs[s][a][tempVal]>=rn) upperbound = tempVal;
                        else lowerbound=tempVal;
                    tempVal=(upperbound+lowerbound)/2 ;
                    }
                    
                    indN[s][a][ns]=(short int)upperbound;
                    }
                    else
                    indN[s][a][ns]=0;
                  
                    
                    //Mdl[s][a][ Samp[s][a][ns] ] += 1/p.sampNum;
                }
            }
     }
        
        clk0.tic();
        Clock clk_try;
        // initializations
        Qf[i][0] = Q;
        vit[0] = 0;
        vcpu[i] = 0.;
        
  
        
        
        

        // main DPP loop
        for (uint j=0; j<=(uint)(p.sampNum); j++) {
            double alpha=1.0/((double)j+1.0);
            for (uint k=0; k<mdp.SSz; k++) {
                for (uint l=0; l<mdp.S[k].ASz; l++) {
                    
                    uint  cindN=(uint)indN[k][l][j];
                    TReward r = mdp.S[k].A[l].r;
                    double max_indN = Q[cindN][0], max_s =  QP[cindN][0];
                    for (uint ap=1; ap<mdp.S[cindN].ASz; ap++) { 
                        if (Q[cindN][ap] > max_indN) max_indN = Q[cindN][ap];
                        if (QP[cindN][ap] > max_s) max_s = QP[cindN][ap];
                    }
 
                    QN[k][l]= (1-alpha)*Q[k][l]+ alpha *r + p.gamma*((1-alpha)*max_indN -(1-2*alpha)*max_s);
                }
            }

            // copy output Q if necessary
            if  (((j%ddt == 0)||(j==(uint)(p.sampNum)-1))&&its<=(uint)(p.dt)) {
                its++;
                Qf[i][its] = QN;
                vit[its] = j;
                vcpu[its*((uint)(p.tryNum)) + i] = clk0.toc();
            }
            
            QP=Q;
            Q=QN;
        }
        mexPrintf("(%d/%d) %.3f scnds\n", i, (uint)p.tryNum, clk_try.toc());
        mexEvalString("drawnow;");
    }
    
    // free random number generator
    gsl_rng_free (r);
}


/* mexFunction is the gateway routine for the MEX-file. */ 
void
mexFunction( int nlhs, mxArray *plhs[],
             int nrhs, const mxArray *prhs[] )
{
  int        i;
  (void) nlhs;     /* unused parameters */
  (void) plhs;

/* Check to see if we are on a platform that does not support the compatibility layer. */
#if defined(_LP64) || defined (_WIN64)
#ifdef MX_COMPAT_32
  for (i=0; i<nrhs; i++)  {
      if (mxIsSparse(prhs[i])) {
          mexErrMsgIdAndTxt("MATLAB:explore:NoSparseCompat",
                    "MEX-files compiled on a 64-bit platform that use sparse array functions need to be compiled using -largeArrayDims.");
      }
  }
#endif
#endif
    if ((nrhs!=2)||(nlhs!=4)) {
        mexPrintf("[QN QO vit vcpu] = dpp1(params,MDP);\n");
        mexPrintf("-------------------------------\n");
        mexPrintf("Arguments:\n\tparams should be struct containing the following fields:\n");
        mexPrintf("\t\ttryNum\n\t\tgamma\n\t\tdt\n\t\ttol\n\t\tsampNum\n\t\trMax\n");
        mexPrintf("\tMDP should be a MDP struct like in 'Model.mat'\n");
        mexPrintf("Output:\n\tQN   : DPP solutions: (ttryNum)x(dt+2) struct with field Q : (NumStates)x(NActions) matrix\n");
        mexPrintf("\tQO   : Optimal solution Q : (NumStates)x(NActions) matrix\n");
        mexPrintf("\tvit  : vector of size (dt+2) with number of iteration\n");
        mexPrintf("\tvcpu : vector of size (dt+2) with cpu times\n");
    }
    else {
        // parse params
        Params p;
        if ( mxIsStruct(prhs[0]) ) {
            mexPrintf("Parsing params\n");
            mexEvalString("drawnow;");
            TReward rMax   = *mxGetPr( mxGetField( prhs[0], 0, "rMax") );
            double sampNum = *mxGetPr( mxGetField( prhs[0], 0, "sampNum") );
            double gamma   = *mxGetPr( mxGetField( prhs[0], 0, "gamma") );
            double tryNum  = *mxGetPr( mxGetField( prhs[0], 0, "tryNum") );
            double tol     = *mxGetPr( mxGetField( prhs[0], 0, "tol") );
            double dt      = *mxGetPr( mxGetField( prhs[0], 0, "dt") );
            p = Params(rMax,sampNum,gamma,tryNum,tol,dt);
        }
        else {
            mexPrintf("Expected params at prhs[0]\n");
        }
        
        // parse model
        MDP mdp;
        if ( mxIsStruct(prhs[1]) ) {
            mexPrintf("Parsing model\n");
            mexEvalString("drawnow;");
            uint SSz(*mxGetPr( mxGetField( prhs[1], 0, "SSz") )); // automatically casts
            mxArray *pS = mxGetField( prhs[1], 0, "S");
    
            TReward r;
            vector<double> P(SSz, (double)0.5); 
            tA Aelem(r,P);
            
            mdp.S.resize(SSz);
            for (uint ns = 0; ns<SSz; ns++) {
                uint ASz(*mxGetPr( mxGetField( pS, ns, "ASz") ));
                vector<tA> a(ASz, Aelem);
                mxArray *A = mxGetField( pS, ns, "A");
                for (uint na = 0; na<ASz; na++) {
                    TReward r = *mxGetPr( mxGetField( A, na, "r" ) );
                    a[na].r = r;
                    mxArray *P = mxGetField( A, na, "P" );
                    mwSize sp = mxGetNumberOfElements(P);
                    double *pP = mxGetPr(P);
                    for (uint si=0; si<sp; si++) {
                        a[na].P[si] = *(pP++);
                    }
                   // mxDestroyArray(P);
                }
                tS s(ASz, a);
                mdp.S[ns] = s;
                //mxDestroyArray(A);
            }
            mdp.SSz = SSz;
            mexPrintf("\n");
           // mxDestroyArray(pS);
        }
        else {
            mexPrintf("Expected model at prhs[1]\n");
        }

        //-----------------------------------------------------------
        //-----------------------------------------------------------
        // DPP 
        tQ Qo;
        tQFinal Qf;
        double * vit, *vcpu;
        vcpu= new double[((uint)p.tryNum)*((uint)(p.dt)+2)];
        vit=new double[(uint)(p.dt)+2];
       
//mexPrintf("\tsize vcpu is %d\n", ((uint)p.tryNum)*((uint)(p.dt)+1));            
        doIt(p, mdp, Qf, Qo, vit, vcpu);

        //-----------------------------------------------------------
        //-----------------------------------------------------------
    
        // build output struct with field Q
        const char *Q[2] = {"Q"};
        plhs[0] = mxCreateStructMatrix((uint)(p.tryNum), (uint)(p.dt)+2, 1,Q);
//mexPrintf("plhs is %d x %d\n", mxGetM(plhs[0]), mxGetN(plhs[0]));
//mexPrintf("Qf is %d x %d\n", Qf.size(), Qf[0].size());
//mexPrintf("Qo\n");
//printQ(Qo);
        uint t=0,r=0;
        for (uint i=0; i<Qf.size()*Qf[0].size(); i++) {
//mexPrintf("t=%d, r=%d, i=%d\n",t,r,i);
//mexPrintf("\taccesing %d \n", Qf[t][r].size());            
//mexPrintf("\taccesing %d x %d\n", Qf[t][r].size(), Qf[t][r][0].size());            
            mxArray *outf = mxCreateNumericMatrix(mdp.SSz, mdp.S[0].ASz, mxDOUBLE_CLASS, mxREAL);
            double *outMatrix1 = mxGetPr(outf);
            uint s=0,a=0;
//mexPrintf("\taccesing %d x %d\n", Qf[t][r].size(), Qf[t][r][0].size());            
            for (uint sa=0; sa<(Qf[t][r].size()*Qf[t][r][0].size()); sa++) {
//mexPrintf("\ts=%d, a=%d, sa=%d\n",s,a,sa);
                outMatrix1[sa]=Qf[t][r][s][a];
                s++;
                if (s==Qf[t][r].size()) {
                    s=0;
                    a++;
 //               a++;
 //               if (a==Qf[t][r][0].size()) {
 //                   a=0;
 //                   s++;
                }
            }
            
            mxSetFieldByNumber(plhs[0], i, 0, outf);
   //         delete [] outMatrix;
            t++;
            if (t==Qf.size()) {
                t=0;
                r++;
            }
        }
    
        // build output optimal Q
        plhs[1] = mxCreateNumericMatrix(mdp.SSz, mdp.S[0].ASz, mxDOUBLE_CLASS, mxREAL);
        double *outMatrix2 = mxGetPr( plhs[1] );
        uint s=0,a=0;
        for (uint sa=0; sa<(Qo.size()*Qo[0].size()); sa++) {
            outMatrix2[sa]=Qo[s][a];
            s++;
            if (s==Qo.size()) {
                s=0;
                a++;
            }
        }

        // timesteps vector
        plhs[2] = mxCreateNumericMatrix(1, (uint)(p.dt)+2, mxDOUBLE_CLASS, mxREAL);
        double *outvit = mxGetPr( plhs[2] );
        memcpy(outvit, vit, sizeof(double)*((uint)(p.dt)+2));

        // cputime vector
        plhs[3] = mxCreateNumericMatrix(p.tryNum, p.dt+2, mxDOUBLE_CLASS, mxREAL);
        double *outvcpu = mxGetPr( plhs[3] );
//mexPrintf("copying\n");
//for (uint i1=0; i1<p.tryNum; i1++)
//	for (uint i2=0; i2<(p.dt+1); i2++)
//		mexPrintf("%d x %d (%d) = %f\n",i1,i2,i1*((uint)(p.dt)+1) + i2,vcpu[i1*((uint)(p.dt)+1) + i2]);
        memcpy(outvcpu, vcpu, sizeof(double)*((uint)p.tryNum)*((uint)(p.dt)+2));
       delete [] vit,vcpu;
    }
    
    return;
}

